@@ -6,6 +6,7 @@ https://botan.randombit.net
 
 (C) 2015,2017,2018,2019,2023 Jack Lloyd
 (C) 2015 Uri  Blumenthal (extensions and patches)
+(C) 2024 Amos Treiber, René Meusel - Rohde & Schwarz Cybersecurity
 
 Botan is released under the Simplified BSD License (see license.txt)
 
@@ -20,13 +21,15 @@ introduced in Botan 3.0.0
 
 from ctypes import CDLL, CFUNCTYPE, POINTER, byref, create_string_buffer, \
     c_void_p, c_size_t, c_uint8, c_uint32, c_uint64, c_int, c_uint, c_char, c_char_p, addressof
+from typing import Callable
 
 from sys import platform
 from time import strptime, mktime, time as system_time
 from binascii import hexlify
 from datetime import datetime
+from collections.abc import Iterable
 
-BOTAN_FFI_VERSION = 20230403
+BOTAN_FFI_VERSION = 20240408
 
 #
 # Base exception for all exceptions raised from this module
@@ -176,7 +179,10 @@ def _set_prototypes(dll):
     ffi_api(dll.botan_cipher_valid_nonce_length, [c_void_p, c_size_t])
     ffi_api(dll.botan_cipher_get_tag_length, [c_void_p, POINTER(c_size_t)])
     ffi_api(dll.botan_cipher_get_default_nonce_length, [c_void_p, POINTER(c_size_t)])
+    ffi_api(dll.botan_cipher_is_authenticated, [c_void_p])
+    ffi_api(dll.botan_cipher_requires_entire_message, [c_void_p])
     ffi_api(dll.botan_cipher_get_update_granularity, [c_void_p, POINTER(c_size_t)])
+    ffi_api(dll.botan_cipher_get_ideal_update_granularity, [c_void_p, POINTER(c_size_t)])
     ffi_api(dll.botan_cipher_query_keylen, [c_void_p, POINTER(c_size_t), POINTER(c_size_t)])
     ffi_api(dll.botan_cipher_get_keyspec, [c_void_p, POINTER(c_size_t), POINTER(c_size_t), POINTER(c_size_t)])
     ffi_api(dll.botan_cipher_set_key, [c_void_p, c_char_p, c_size_t])
@@ -278,6 +284,7 @@ def _set_prototypes(dll):
 
     ffi_api(dll.botan_privkey_view_der, [c_void_p, c_void_p, VIEW_BIN_CALLBACK])
     ffi_api(dll.botan_privkey_view_pem, [c_void_p, c_void_p, VIEW_STR_CALLBACK])
+    ffi_api(dll.botan_privkey_view_raw, [c_void_p, c_void_p, VIEW_BIN_CALLBACK])
 
     ffi_api(dll.botan_privkey_algo_name, [c_void_p, c_char_p, POINTER(c_size_t)])
     ffi_api(dll.botan_privkey_export_encrypted,
@@ -303,6 +310,7 @@ def _set_prototypes(dll):
 
     ffi_api(dll.botan_pubkey_view_der, [c_void_p, c_void_p, VIEW_BIN_CALLBACK])
     ffi_api(dll.botan_pubkey_view_pem, [c_void_p, c_void_p, VIEW_STR_CALLBACK])
+    ffi_api(dll.botan_pubkey_view_raw, [c_void_p, c_void_p, VIEW_BIN_CALLBACK])
 
     ffi_api(dll.botan_pubkey_algo_name, [c_void_p, c_char_p, POINTER(c_size_t)])
     ffi_api(dll.botan_pubkey_check_key, [c_void_p, c_void_p, c_uint32], [-1])
@@ -339,14 +347,30 @@ def _set_prototypes(dll):
     ffi_api(dll.botan_pubkey_load_ed25519, [c_void_p, c_char_p])
     ffi_api(dll.botan_privkey_ed25519_get_privkey, [c_void_p, c_char_p])
     ffi_api(dll.botan_pubkey_ed25519_get_pubkey, [c_void_p, c_char_p])
+    ffi_api(dll.botan_privkey_load_ed448, [c_void_p, c_char_p])
+    ffi_api(dll.botan_pubkey_load_ed448, [c_void_p, c_char_p])
+    ffi_api(dll.botan_privkey_ed448_get_privkey, [c_void_p, c_char_p])
+    ffi_api(dll.botan_pubkey_ed448_get_pubkey, [c_void_p, c_char_p])
     ffi_api(dll.botan_privkey_load_x25519, [c_void_p, c_char_p])
     ffi_api(dll.botan_pubkey_load_x25519, [c_void_p, c_char_p])
     ffi_api(dll.botan_privkey_x25519_get_privkey, [c_void_p, c_char_p])
     ffi_api(dll.botan_pubkey_x25519_get_pubkey, [c_void_p, c_char_p])
-    ffi_api(dll.botan_privkey_load_kyber, [c_void_p, c_char_p])
-    ffi_api(dll.botan_pubkey_load_kyber, [c_void_p, c_char_p])
+    ffi_api(dll.botan_privkey_load_x448, [c_void_p, c_char_p])
+    ffi_api(dll.botan_pubkey_load_x448, [c_void_p, c_char_p])
+    ffi_api(dll.botan_privkey_x448_get_privkey, [c_void_p, c_char_p])
+    ffi_api(dll.botan_pubkey_x448_get_pubkey, [c_void_p, c_char_p])
+    ffi_api(dll.botan_privkey_load_ml_dsa, [c_void_p, c_void_p, c_int, c_char_p])
+    ffi_api(dll.botan_pubkey_load_ml_dsa, [c_void_p, c_void_p, c_int, c_char_p])
+    ffi_api(dll.botan_privkey_load_slh_dsa, [c_void_p, c_void_p, c_int, c_char_p])
+    ffi_api(dll.botan_pubkey_load_slh_dsa, [c_void_p, c_void_p, c_int, c_char_p])
+    ffi_api(dll.botan_privkey_load_kyber, [c_void_p, c_char_p, c_int])
+    ffi_api(dll.botan_pubkey_load_kyber, [c_void_p, c_char_p, c_int])
     ffi_api(dll.botan_privkey_view_kyber_raw_key, [c_void_p, c_void_p, VIEW_BIN_CALLBACK])
     ffi_api(dll.botan_pubkey_view_kyber_raw_key, [c_void_p, c_void_p, VIEW_BIN_CALLBACK])
+    ffi_api(dll.botan_privkey_load_ml_kem, [c_void_p, c_void_p, c_int, c_char_p])
+    ffi_api(dll.botan_pubkey_load_ml_kem, [c_void_p, c_void_p, c_int, c_char_p])
+    ffi_api(dll.botan_privkey_load_frodokem, [c_void_p, c_void_p, c_int, c_char_p])
+    ffi_api(dll.botan_pubkey_load_frodokem, [c_void_p, c_void_p, c_int, c_char_p])
     ffi_api(dll.botan_privkey_load_ecdsa, [c_void_p, c_void_p, c_char_p])
     ffi_api(dll.botan_pubkey_load_ecdsa, [c_void_p, c_void_p, c_void_p, c_char_p])
     ffi_api(dll.botan_pubkey_load_ecdh, [c_void_p, c_void_p, c_void_p, c_char_p])
@@ -491,6 +515,16 @@ def _set_prototypes(dll):
     ffi_api(dll.botan_zfec_decode,
             [c_size_t, c_size_t, POINTER(c_size_t), POINTER(c_char_p), c_size_t, POINTER(c_char_p)])
 
+    # TPM2
+    ffi_api(dll.botan_tpm2_supports_crypto_backend, [])
+    ffi_api(dll.botan_tpm2_ctx_init, [c_void_p, c_char_p], [-40])
+    ffi_api(dll.botan_tpm2_ctx_init_ex, [c_void_p, c_char_p, c_char_p], [-40])
+    ffi_api(dll.botan_tpm2_ctx_enable_crypto_backend, [c_void_p, c_void_p])
+    ffi_api(dll.botan_tpm2_ctx_destroy, [c_void_p], [-40])
+    ffi_api(dll.botan_tpm2_rng_init, [c_void_p, c_void_p, c_void_p, c_void_p, c_void_p])
+    ffi_api(dll.botan_tpm2_unauthenticated_session_init, [c_void_p, c_void_p])
+    ffi_api(dll.botan_tpm2_session_destroy, [c_void_p])
+
     return dll
 
 #
@@ -620,14 +654,115 @@ def const_time_compare(x, y):
     rc = _DLL.botan_constant_time_compare(xbits, ybits, c_size_t(len_x))
     return rc == 0
 
+#
+# TPM2
+#
+
+class TPM2Object:
+    def __init__(self, obj: c_void_p, destroyer: Callable[[c_void_p], None]):
+        self.__obj = obj
+        self.__destroyer = destroyer
+
+    def __del__(self):
+        if hasattr(self, '__obj') and hasattr(self, '__destroyer'):
+            self.__destroyer(self.__obj)
+
+    def handle_(self):
+        return self.__obj
+
+class TPM2Context(TPM2Object):
+    """TPM 2.0 Context object"""
+
+    def __init__(self, tcti_name_maybe_with_conf: str = None, tcti_conf: str = None):
+        """Construct a TPM2Context object with optional TCTI name and configuration."""
+
+        obj = c_void_p(0)
+        if tcti_conf is not None:
+            rc = _DLL.botan_tpm2_ctx_init_ex(byref(obj), _ctype_str(tcti_name_maybe_with_conf), _ctype_str(tcti_conf))
+        else:
+            rc = _DLL.botan_tpm2_ctx_init(byref(obj), _ctype_str(tcti_name_maybe_with_conf))
+        if rc == -40: # 'Not Implemented'
+            raise BotanException("TPM2 is not implemented in this build configuration", rc)
+        self.rng_ = None
+        super().__init__(obj, _DLL.botan_tpm2_ctx_destroy)
+
+    @staticmethod
+    def supports_botan_crypto_backend() -> bool:
+        """Returns True if the given build supports the Botan-based crypto backend."""
+        rc = _DLL.botan_tpm2_supports_crypto_backend()
+        return rc == 1
+
+    def enable_botan_crypto_backend(self, rng):
+        """Enables the Botan-based crypto backend.
+        The passed rng MUST NOT be dependent on the TPM."""
+        # By keeping a reference to the passed-in RNG object, we make sure
+        # that the underlying object lives at least as long as this context.
+        self.rng_ = rng
+        _DLL.botan_tpm2_ctx_enable_crypto_backend(self.handle_(), self.rng_.handle_())
+
+class TPM2Session(TPM2Object):
+    """Basic TPM 2.0 Session object, typically users will instantiate a derived class."""
+
+    def __init__(self, obj: c_void_p):
+        super().__init__(obj, _DLL.botan_tpm2_session_destroy)
+
+    @staticmethod
+    def session_bundle_(*args):
+        """Transforms a session bundle passed by the downstream user into a 3-tuple of session handles.
+        Users might pass a bare TPM2Session object or an iterable list of such objects."""
+        if len(args) == 1:
+            if isinstance(args[0], Iterable):
+                args = list(args[0])
+            elif args[0] is None:
+                args = []
+
+        if len(args) <= 3 and all(isinstance(s, TPM2Session) for s in args):
+            sessions = list(args)
+            while len(sessions) < 3:
+                sessions.append(None)
+            return (s.handle_() if isinstance(s, TPM2Session) else None for s in sessions)
+        else:
+            raise BotanException("session bundle arguments must be 0 to 3 TPM2Session objects")
+
+
+class TPM2UnauthenticatedSession(TPM2Session):
+    """Session object that is not bound to any authenication credential.
+    It provides basic parameter encryption between the application and the TPM."""
+    def __init__(self, ctx: TPM2Context):
+        obj = c_void_p(0)
+        _DLL.botan_tpm2_unauthenticated_session_init(byref(obj), ctx.handle_())
+        super().__init__(obj)
+
 #
 # RNG
 #
 class RandomNumberGenerator:
     # Can also use type "system"
-    def __init__(self, rng_type='system'):
+    def __init__(self, rng_type='system', **kwargs):
+        """Constructs a RandomNumberGenerator of type rng_type
+
+        Available RNG types are::
+
+        * 'system': Adapter to the operating system's RNG
+        * 'user':   Software-PRNG that is auto-seeded by the system RNG
+        * 'null':   Mock-RNG that fails if randomness is pulled from it
+        * 'hwrng':  Adapter to an available hardware RNG (platform dependent)
+        * 'tpm2':   Adapter to a TPM 2.0 RNG
+                    (needs additional named arguments tpm2_context= and, optionally, tpm2_sessions=)
+        """
         self.__obj = c_void_p(0)
-        _DLL.botan_rng_init(byref(self.__obj), _ctype_str(rng_type))
+        if rng_type == 'tpm2':
+            ctx = kwargs.pop("tpm2_context", None)
+            if not ctx or not isinstance(ctx, TPM2Context):
+                raise BotanException("Cannot instantiate a TPM2-based RNG without a TPM2 context, pass tpm2_context= argument?")
+            sessions = TPM2Session.session_bundle_(kwargs.pop("tpm2_sessions", None))
+            if kwargs:
+                raise BotanException("Unexpected arguments for TPM2 RNG: %s" % (", ".join(kwargs.keys())))
+            _DLL.botan_tpm2_rng_init(byref(self.__obj), ctx.handle_(), *sessions)
+        else:
+            if kwargs:
+                raise BotanException("Unexpected arguments for RNG type %s: %s" % (rng_type, ", ".join(kwargs.keys())))
+            _DLL.botan_rng_init(byref(self.__obj), _ctype_str(rng_type))
 
     def __del__(self):
         _DLL.botan_rng_destroy(self.__obj)
@@ -845,6 +980,11 @@ class SymmetricCipher:
         _DLL.botan_cipher_get_update_granularity(self.__obj, byref(l))
         return l.value
 
+    def ideal_update_granularity(self):
+        l = c_size_t(0)
+        _DLL.botan_cipher_get_ideal_update_granularity(self.__obj, byref(l))
+        return l.value
+
     def key_length(self):
         kmin = c_size_t(0)
         kmax = c_size_t(0)
@@ -867,7 +1007,8 @@ class SymmetricCipher:
         return l.value
 
     def is_authenticated(self):
-        return self.tag_length() > 0
+        rc = _DLL.botan_cipher_is_authenticated(self.__obj)
+        return rc == 1
 
     def valid_nonce_length(self, nonce_len):
         rc = _DLL.botan_cipher_valid_nonce_length(self.__obj, nonce_len)
@@ -1099,6 +1240,30 @@ class PublicKey: # pylint: disable=invalid-name
         _DLL.botan_pubkey_load_kyber(byref(obj), key, len(key))
         return PublicKey(obj)
 
+    @classmethod
+    def load_ml_kem(cls, mlkem_mode, key):
+        obj = c_void_p(0)
+        _DLL.botan_pubkey_load_ml_kem(byref(obj), key, len(key), _ctype_str(mlkem_mode))
+        return PublicKey(obj)
+
+    @classmethod
+    def load_ml_dsa(cls, mldsa_mode, key):
+        obj = c_void_p(0)
+        _DLL.botan_pubkey_load_ml_dsa(byref(obj), key, len(key), _ctype_str(mldsa_mode))
+        return PublicKey(obj)
+
+    @classmethod
+    def load_slh_dsa(cls, slhdsa_mode, key):
+        obj = c_void_p(0)
+        _DLL.botan_pubkey_load_slh_dsa(byref(obj), key, len(key), _ctype_str(slhdsa_mode))
+        return PublicKey(obj)
+
+    @classmethod
+    def load_frodokem(cls, frodo_mode, key):
+        obj = c_void_p(0)
+        _DLL.botan_pubkey_load_frodokem(byref(obj), key, len(key), _ctype_str(frodo_mode))
+        return PublicKey(obj)
+
     def __del__(self):
         _DLL.botan_pubkey_destroy(self.__obj)
 
@@ -1130,7 +1295,11 @@ class PublicKey: # pylint: disable=invalid-name
     def to_pem(self):
         return _call_fn_viewing_str(lambda vc, vfn: _DLL.botan_pubkey_view_pem(self.__obj, vc, vfn))
 
+    def to_raw(self):
+        return _call_fn_viewing_vec(lambda vc, vfn: _DLL.botan_pubkey_view_raw(self.__obj, vc, vfn))
+
     def view_kyber_raw_key(self):
+        """Deprecated: use to_raw() instead"""
         return _call_fn_viewing_vec(lambda vc, vfn: _DLL.botan_pubkey_view_kyber_raw_key(self.__obj, vc, vfn))
 
     def fingerprint(self, hash_algorithm='SHA-256'):
@@ -1173,8 +1342,11 @@ class PrivateKey:
         elif algo == 'ecdsa':
             algo = 'ECDSA'
         elif algo in ['ecdh', 'ECDH']:
-            if params == 'curve25519':
-                algo = 'Curve25519'
+            if params in ['x25519', 'curve25519']:
+                algo = 'X25519'
+                params = ''
+            elif params == 'x448':
+                algo = 'X448'
                 params = ''
             else:
                 algo = 'ECDH'
@@ -1251,6 +1423,30 @@ class PrivateKey:
         _DLL.botan_privkey_load_kyber(byref(obj), key, len(key))
         return PrivateKey(obj)
 
+    @classmethod
+    def load_ml_kem(cls, mlkem_mode, key):
+        obj = c_void_p(0)
+        _DLL.botan_privkey_load_ml_kem(byref(obj), key, len(key), _ctype_str(mlkem_mode))
+        return PrivateKey(obj)
+
+    @classmethod
+    def load_ml_dsa(cls, mldsa_mode, key):
+        obj = c_void_p(0)
+        _DLL.botan_privkey_load_ml_dsa(byref(obj), key, len(key), _ctype_str(mldsa_mode))
+        return PrivateKey(obj)
+
+    @classmethod
+    def load_slh_dsa(cls, slh_dsa, key):
+        obj = c_void_p(0)
+        _DLL.botan_privkey_load_slh_dsa(byref(obj), key, len(key), _ctype_str(slh_dsa))
+        return PrivateKey(obj)
+
+    @classmethod
+    def load_frodokem(cls, frodo_mode, key):
+        obj = c_void_p(0)
+        _DLL.botan_privkey_load_frodokem(byref(obj), key, len(key), _ctype_str(frodo_mode))
+        return PrivateKey(obj)
+
     def __del__(self):
         _DLL.botan_privkey_destroy(self.__obj)
 
@@ -1276,7 +1472,11 @@ class PrivateKey:
     def to_pem(self):
         return _call_fn_viewing_str(lambda vc, vfn: _DLL.botan_privkey_view_pem(self.__obj, vc, vfn))
 
+    def to_raw(self):
+        return _call_fn_viewing_vec(lambda vc, vfn: _DLL.botan_privkey_view_raw(self.__obj, vc, vfn))
+
     def view_kyber_raw_key(self):
+        """Deprecated: use to_raw() instead"""
         return _call_fn_viewing_vec(lambda vc, vfn: _DLL.botan_privkey_view_kyber_raw_key(self.__obj, vc, vfn))
 
     def export(self, pem=False):
@@ -1493,6 +1693,7 @@ def _load_buf_or_file(filename, buf, file_fn, buf_fn):
 #
 class X509Cert: # pylint: disable=invalid-name
     def __init__(self, filename=None, buf=None):
+        self.__obj = c_void_p(0)
         self.__obj = _load_buf_or_file(filename, buf, _DLL.botan_x509_cert_load_file, _DLL.botan_x509_cert_load)
 
     def __del__(self):
@@ -1671,6 +1872,7 @@ class X509Cert: # pylint: disable=invalid-name
 #
 class X509CRL:
     def __init__(self, filename=None, buf=None):
+        self.__obj = c_void_p(0)
         self.__obj = _load_buf_or_file(filename, buf, _DLL.botan_x509_crl_load_file, _DLL.botan_x509_crl_load)
 
     def __del__(self):
